import numpy as np

from ...types import *
from ....types import *
from ..base import Converter


@Converter.register("string")
class StringConverter(Converter):
    def _fit(self) -> "Converter":
        assert self.info.is_valid and self.info.is_string and self.info.need_transform
        self._transform_dict = self._recognizer.transform_dict
        self._reverse_transform_dict = dict(map(reversed, self._transform_dict.items()))
        return self

    def _convert(self,
                 flat_arr: flat_arr_type) -> np.ndarray:
        return np.array([self._transform_dict.get(elem, 0) for elem in flat_arr], np_float_type)

    def _recover(self,
                 flat_arr: flat_arr_type) -> np.ndarray:
        return np.array([self._reverse_transform_dict[elem] for elem in flat_arr], np.str)


__all__ = ["StringConverter"]



import numpy as np

from ..base import Processor


@Processor.register("normalize")
class Normalize(Processor):
    @property
    def input_dim(self) -> int:
        return 1

    @property
    def output_dim(self) -> int:
        return 1

    def _initialize(self, **kwargs) -> None:
        self._eps = kwargs.get("eps", 1e-8)

    def fit(self,
            columns: np.ndarray) -> Processor:
        self._caches["mean"], self._caches["std"] = columns.mean(), columns.std() + self._eps
        return self

    def _process(self,
                 columns: np.ndarray) -> np.ndarray:
        mean, std = map(self._caches.get, ["mean", "std"])
        columns -= mean
        columns /= std
        return columns

    def _recover(self,
                 processed_columns: np.ndarray) -> np.ndarray:
        mean, std = map(self._caches.get, ["mean", "std"])
        processed_columns *= std
        processed_columns += mean
        return processed_columns


__all__ = ["Normalize"]



import numpy as np

from ..base import Processor


@Processor.register("lambda")
class Lambda(Processor):
    def _initialize(self, **kwargs):
        self._process = kwargs["process"]
        self._recover = kwargs["recover"]
        self._input_dim = kwargs["input_dim"]
        self._output_dim = kwargs["output_dim"]

    @property
    def input_dim(self) -> int:
        return self._input_dim

    @property
    def output_dim(self) -> int:
        return self._output_dim

    def fit(self,
            columns: np.ndarray) -> Processor:
        return self

    def _process(self,
                 columns: np.ndarray) -> np.ndarray:
        return self._process(columns)

    def _recover(self,
                 processed_columns: np.ndarray) -> np.ndarray:
        return self._recover(processed_columns)


__all__ = ["Lambda"]



import numpy as np

from typing import *

from ...types import *
from ....types import *
from ..base import Converter


@Converter.register("numerical")
class NumericalConverter(Converter):
    @property
    def nan_fill(self) -> Union[float, None]:
        return None if self._nan_fill is None else self._feature_statistics[self._nan_fill]

    @property
    def statistics(self) -> Dict[str, float]:
        return self._feature_statistics

    def _initialize(self, **kwargs) -> None:
        self._nan_fill = kwargs.get("nan_fill", "median")
        self._feature_statistics = {}

    def _fit(self) -> "Converter":
        assert self.info.is_valid and self.info.is_numerical
        nan_mask = self.info.nan_mask
        np_flat_features = self.info.flat_arr.copy()
        base_attrs = ["median", "mean", "std", "min", "max"]
        for attr in base_attrs:
            if nan_mask is None:
                np_flat_valid = np_flat_features
            else:
                np_flat_valid = np_flat_features[~nan_mask]
            self._feature_statistics[attr] = getattr(np, attr)(np_flat_valid).item()
        if self._nan_fill is None or not self.info.contains_nan:
            self._converted_features = np_flat_features
        else:
            np_flat_features[nan_mask] = self.nan_fill
            self._converted_features = np_flat_features
        return self

    def _convert(self,
                 flat_arr: flat_arr_type) -> np.ndarray:
        np_flat = np.asarray(flat_arr, np_float_type)
        if self._nan_fill is None:
            return np_flat
        np_flat[np.isnan(np_flat)] = self.nan_fill
        return np_flat

    def _recover(self,
                 flat_arr: flat_arr_type) -> np.ndarray:
        return np.asarray(flat_arr, np_float_type)


__all__ = ["NumericalConverter"]



import numpy as np

from sklearn.preprocessing import OneHotEncoder

from ..base import Processor
from ....types import np_float_type


@Processor.register("one_hot")
class OneHot(Processor):
    def _initialize(self, **kwargs) -> None:
        self._categories = kwargs.get("categories", "auto")

    @property
    def input_dim(self) -> int:
        return 1

    @property
    def output_dim(self) -> int:
        return len(self._all_categories)

    def fit(self,
            columns: np.ndarray) -> Processor:
        self._encoder = OneHotEncoder(categories=self._categories, sparse=False, dtype=np_float_type)
        self._encoder.fit(columns)
        self._all_categories = self._encoder.categories_[0]
        return self

    def _process(self,
                 columns: np.ndarray) -> np.ndarray:
        return self._encoder.transform(columns)

    def _recover(self,
                 processed_columns: np.ndarray) -> np.ndarray:
        return self._encoder.inverse_transform(processed_columns)


__all__ = ["OneHot"]



import numpy as np

from typing import *
from functools import partial
from cftool.misc import *

from .types import *
from ..types import *


class SplitResult(NamedTuple):
    dataset: TabularDataset
    corresponding_indices: np.ndarray
    remaining_indices: np.ndarray

    @classmethod
    def concat(cls,
               results: List["SplitResult"]) -> "SplitResult":
        datasets = [result.dataset for result in results]
        basic_info = datasets[0][2:]
        x_list, y_list = zip(*[dataset.xy for dataset in datasets])
        x_concat, y_concat = map(np.vstack, [x_list, y_list])
        return SplitResult(
            TabularDataset(x_concat, y_concat, *basic_info),
            np.hstack([result.corresponding_indices for result in results]),
            np.hstack([result.remaining_indices for result in results])
        )


class DataSplitter(SavingMixin):
    """
    Util class for dividing dataset based on task type
    * If it's regression task, it's simple to split data
    * If it's classification task, we need to split data based on labels, because we need
    to ensure the divided data contain all labels available

    Examples
    --------
    >>> import numpy as np
    >>>
    >>> from cfdata.types import np_int_type
    >>> from cfdata.tabular.types import TaskTypes
    >>> from cfdata.tabular.wrapper import TabularDataset
    >>> from cfdata.tabular.utils import DataSplitter
    >>>
    >>> x = np.arange(12).reshape([6, 2])
    >>> # create an imbalance dataset
    >>> y = np.zeros(6, np_int_type)
    >>> y[[-1, -2]] = 1
    >>> dataset = TabularDataset.from_xy(x, y, TaskTypes.CLASSIFICATION)
    >>> data_splitter = DataSplitter().fit(dataset)
    >>> # labels in result will keep its ratio
    >>> result = data_splitter.split(3)
    >>> # [0 0 1]
    >>> print(result.dataset.y.ravel())
    >>> data_splitter.reset()
    >>> result = data_splitter.split(0.5)
    >>> # [0 0 1]
    >>> print(result.dataset.y.ravel())
    >>> # at least one sample of each class will be kept
    >>> y[-2] = 0
    >>> dataset = TabularDataset.from_xy(x, y, TaskTypes.CLASSIFICATION)
    >>> data_splitter = DataSplitter().fit(dataset)
    >>> result = data_splitter.split(2)
    >>> # [0 0 0 0 0 1] [0 1]
    >>> print(y, result.dataset.y.ravel())

    """

    @property
    def data_tuple_base(self) -> Union[None, Type[NamedTuple]]:
        return

    @property
    def data_tuple_attributes(self) -> Union[None, List[str]]:
        return

    def __init__(self,
                 *,
                 time_series_config: dict = None,
                 shuffle: bool = True,
                 replace: bool = False,
                 verbose_level: int = 2):
        self._remained_indices = None
        self._time_indices_list = self._time_indices_list_in_use = None
        self._label_indices_list = self._label_indices_list_in_use = None
        self._time_series_config, self._time_series_sorting_indices = time_series_config, None
        self._shuffle, self._replace = shuffle, replace
        self._verbose_level = verbose_level
        if time_series_config is not None:
            if replace:
                raise ValueError("`replace` cannot be True when splitting time series dataset")
            self._id_column_setting = time_series_config.get("id_column")
            self._time_column_setting = time_series_config.get("time_column")
            if self._id_column_setting is None:
                raise ValueError("id_column should be provided in time_series_config")
            if self._time_column_setting is None:
                raise ValueError("time_column should be provided in time_series_config")
            self._id_column_is_int, self._time_column_is_int = map(
                lambda column: isinstance(column, int), [self._id_column_setting, self._time_column_setting])

    @property
    def x(self) -> np.ndarray:
        return self._x

    @property
    def y(self) -> np.ndarray:
        return self._y

    @property
    def id_column(self):
        return self._id_column

    @property
    def time_column(self):
        return self._time_column

    @property
    def is_time_series(self):
        return self._time_series_config is not None

    @property
    def sorting_indices(self):
        if not self.is_time_series:
            raise ValueError("sorting_indices should not be called when it is not time series condition")
        return self._time_series_sorting_indices

    @property
    def remained_indices(self):
        return self._remained_indices[::-1].copy()

    @property
    def remained_xy(self):
        indices = self.remained_indices
        return self._x[indices], self._y[indices]

    # reset methods

    def _reset_reg(self):
        n_data = len(self._x)
        if not self._shuffle:
            self._remained_indices = np.arange(n_data)
        else:
            self._remained_indices = np.random.permutation(n_data)
        self._remained_indices = self._remained_indices.astype(np_int_type)

    def _reset_clf(self):
        if self._label_indices_list is None:
            flattened_y = self._y.ravel()
            unique_indices = get_unique_indices(flattened_y)
            self._unique_labels, counts = unique_indices[:2]
            self._label_indices_list = unique_indices.split_indices
            self._n_samples = len(flattened_y)
            self._label_ratios = counts / self._n_samples
            self._n_unique_labels = len(self._unique_labels)
            if self._n_unique_labels == 1:
                raise ValueError("only 1 unique label is detected, which is invalid in classification task")
            self._unique_labels = self._unique_labels.astype(np_int_type)
            self._label_indices_list = list(map(partial(np.asarray, dtype=np_int_type), self._label_indices_list))
        self._reset_indices_list("label_indices_list")

    def _reset_time_series(self):
        if self._time_indices_list is None:
            self.log_msg(f"gathering time -> indices mapping", self.info_prefix, verbose_level=5)
            self._unique_times, times_counts, self._time_indices_list = map(
                lambda arr: arr[::-1],
                get_unique_indices(self._time_column)
            )
            self._times_counts_cumsum = np.cumsum(times_counts).astype(np_int_type)
            assert self._times_counts_cumsum[-1] == len(self._time_column)
            self._time_series_sorting_indices = np.hstack(self._time_indices_list[::-1]).astype(np_int_type)
            self._unique_times = self._unique_times.astype(np_int_type)
            self._time_indices_list = list(map(partial(np.asarray, dtype=np_int_type), self._time_indices_list))
        self._reset_indices_list("time_indices_list")
        self._times_counts_cumsum_in_use = self._times_counts_cumsum.copy()

    def _reset_indices_list(self, attr):
        self_attr = getattr(self, f"_{attr}")
        if self._shuffle:
            tuple(map(np.random.shuffle, self_attr))
        attr_in_use = f"_{attr}_in_use"
        setattr(self, attr_in_use, [arr.copy() for arr in self_attr])
        self._remained_indices = np.hstack(getattr(self, attr_in_use)).astype(np_int_type)

    # split methods

    def _split_reg(self, n: int):
        tgt_indices = self._remained_indices[-n:]
        n = min(n, len(self._remained_indices) - 1)
        if self._replace:
            np.random.shuffle(self._remained_indices)
        elif n > 0:
            self._remained_indices = self._remained_indices[:-n]
        return tgt_indices

    def _split_clf(self, n: int):
        if n < self._n_unique_labels:
            raise ValueError(
                f"at least {self._n_unique_labels} samples are required because "
                f"we have {self._n_unique_labels} unique labels"
            )
        pop_indices_list, tgt_indices_list = [], []
        n_samples_per_label = np.maximum(1, np.round(n * self._label_ratios).astype(np_int_type))
        # -n_unique_labels <= n_samples_exceeded <= n_unique_labels
        n_samples_exceeded = n_samples_per_label.sum() - n
        # adjust n_samples_per_label to make sure `n` samples are split out
        if n_samples_exceeded != 0:
            sign, n_samples_exceeded = np.sign(n_samples_exceeded), abs(n_samples_exceeded)
            chosen_indices = np.arange(self._n_unique_labels)[n_samples_per_label != 1]
            np.random.shuffle(chosen_indices)
            n_chosen_indices = len(chosen_indices)
            n_tile = int(np.ceil(n_samples_exceeded / n_chosen_indices))
            n_proceeded = 0
            for _ in range(n_tile - 1):
                n_samples_per_label[chosen_indices] -= sign
                n_proceeded += n_chosen_indices
            for idx in chosen_indices[:n_samples_exceeded - n_proceeded]:
                n_samples_per_label[idx] -= sign
        assert n_samples_per_label.sum() == n
        n_overlap = 0
        for indices, n_sample_per_label in zip(self._label_indices_list_in_use, n_samples_per_label):
            n_samples_in_use = len(indices)
            tgt_indices_list.append(indices[-n_sample_per_label:])
            if n_sample_per_label >= n_samples_in_use:
                pop_indices_list.append([])
                n_overlap += n_sample_per_label
            else:
                pop_indices_list.append(np.arange(n_samples_in_use - n_sample_per_label, n_samples_in_use))
        tgt_indices = np.hstack(tgt_indices_list)
        if self._replace:
            tuple(map(np.random.shuffle, self._label_indices_list_in_use))
            self._remained_indices = np.hstack(self._label_indices_list_in_use)
        else:
            self._label_indices_list_in_use = list(map(
                lambda arr, pop_indices: np.delete(arr, pop_indices),
                self._label_indices_list_in_use, pop_indices_list
            ))
            remain_indices = np.hstack(self._label_indices_list_in_use)
            base = np.zeros(self._n_samples)
            base[tgt_indices] += 1
            base[remain_indices] += 1
            assert np.sum(base >= 2) <= n_overlap
            self._remained_indices = remain_indices
        return tgt_indices

    def _split_time_series(self, n: int):
        split_arg = np.argmax(self._times_counts_cumsum_in_use >= n)
        n_left = self._times_counts_cumsum_in_use[split_arg] - n
        if split_arg == 0:
            n_res, selected_indices = n, []
        else:
            n_res = n - self._times_counts_cumsum_in_use[split_arg - 1]
            selected_indices = self._time_indices_list_in_use[:split_arg]
            self._time_indices_list_in_use = self._time_indices_list_in_use[split_arg:]
            self._times_counts_cumsum_in_use = self._times_counts_cumsum_in_use[split_arg:]
        selected_indices.append(self._time_indices_list_in_use[0][:n_res])
        if n_left > 0:
            self._time_indices_list_in_use[0] = self._time_indices_list_in_use[0][n_res:]
        else:
            self._time_indices_list_in_use = self._time_indices_list_in_use[1:]
            self._times_counts_cumsum_in_use = self._times_counts_cumsum_in_use[1:]
        tgt_indices, self._remained_indices = map(np.hstack, [selected_indices, self._time_indices_list_in_use])
        self._times_counts_cumsum_in_use -= n
        return tgt_indices[::-1].copy()

    def fit(self,
            dataset: TabularDataset) -> "DataSplitter":
        self._dataset = dataset
        self._is_regression = dataset.is_reg
        self._x = dataset.x
        self._y = dataset.y
        if not self.is_time_series:
            self._time_column = None
        else:
            if not self._id_column_is_int and not self._time_column_is_int:
                self._id_column, self._time_column = map(
                    np.asarray, [self._id_column_setting, self._time_column_setting])
            else:
                id_column, time_column = self._id_column_setting, self._time_column_setting
                error_msg_prefix = "id_column & time_column should both be int, but"
                if not self._id_column_is_int:
                    raise ValueError(f"{error_msg_prefix} id_column='{id_column}' found")
                if not self._time_column_is_int:
                    raise ValueError(f"{error_msg_prefix} time_column='{time_column}' found")
                if id_column < time_column:
                    id_first = True
                    split_list = [id_column, id_column + 1, time_column, time_column + 1]
                else:
                    id_first = False
                    split_list = [time_column, time_column + 1, id_column, id_column + 1]
                columns = np.split(self._x, split_list, axis=1)
                if id_first:
                    self._id_column, self._time_column = columns[1], columns[3]
                else:
                    self._id_column, self._time_column = columns[3], columns[1]
                self._x = np.hstack([columns[0], columns[2], columns[4]])
            self._id_column, self._time_column = map(np.ravel, [self._id_column, self._time_column])
        return self.reset()

    def reset(self) -> "DataSplitter":
        if self._time_column is not None:
            self._reset_time_series()
        elif self._is_regression:
            self._reset_reg()
        else:
            self._reset_clf()
        return self

    def split(self,
              n: Union[int, float]) -> SplitResult:
        error_msg = "please call 'reset' method before calling 'split' method"
        if self._is_regression and self._remained_indices is None:
            raise ValueError(error_msg)
        if not self._is_regression and self._label_indices_list_in_use is None:
            raise ValueError(error_msg)
        if n >= len(self._remained_indices):
            remained_x, remained_y = self.remained_xy
            return SplitResult(
                TabularDataset.from_xy(remained_x, remained_y, self._dataset.task_type),
                self._remained_indices, np.array([], np.int)
            )
        if n < 1. or (n == 1. and isinstance(n, float)):
            n = int(round(len(self._x) * n))
        if self.is_time_series:
            split_method = self._split_time_series
        else:
            split_method = self._split_reg if self._is_regression else self._split_clf
        tgt_indices = split_method(n)
        assert len(tgt_indices) == n
        x_split, y_split = self._x[tgt_indices], self._y[tgt_indices]
        dataset_split = TabularDataset(x_split, y_split, *self._dataset[2:])
        return SplitResult(dataset_split, tgt_indices, self._remained_indices)

    def split_multiple(self,
                       n_list: List[Union[int, float]],
                       *,
                       return_remained: bool = False) -> List[SplitResult]:
        n_list = n_list.copy()
        n_total = len(self._x)
        if not all(n_ <= 1. for n_ in n_list):
            if any(n_ < 1. for n_ in n_list):
                raise ValueError("some of the elements in `n_list` (but not all) are less than 1")
            if return_remained:
                n_list.append(n_total - sum(n_list))
        else:
            ratio_sum = sum(n_list)
            if ratio_sum > 1.:
                raise ValueError("sum of `n_list` should not be greater than 1")
            if return_remained and ratio_sum == 1:
                raise ValueError("sum of `n_list` should be less than 1 "
                                 "when `return_remained` is True")
            n_selected = int(round(n_total * ratio_sum))
            n_list[:-1] = [int(round(n_total * ratio)) for ratio in n_list[:-1]]
            n_list[-1] = n_selected - sum(n_list[:-1])
            if ratio_sum < 1.:
                n_list.append(n_total - n_selected)
        return list(map(self.split, n_list))


class KFold:
    """
    Util class which can perform k-fold data splitting:

    1. X = {x1, x2, ..., xn} -> [X1, X2, ..., Xk]
    2. Xi ∩ Xj = ∅, ∀ i, j = 1,..., K
    3. X1 ∪ X2 ∪ ... ∪ Xk = X

    * Notice that `KFold` does not always hold the principles listed above, because `DataSplitter`
    will ensure that at least one sample of each class will be kept. In this case, when we apply
    `KFold` to an imbalance dataset, `KFold` may slightly violate principle 2. and 3.

    Parameters
    ----------
    k : int, number of folds
    dataset : TabularDataset, dataset which we want to split
    **kwargs : used to initialize `DataSplitter` instance

    Examples
    ----------
    >>> import numpy as np
    >>>
    >>> from cfdata.types import np_int_type
    >>> from cfdata.tabular.types import TaskTypes
    >>> from cfdata.tabular.wrapper import TabularDataset
    >>> from cfdata.tabular.utils import KFold
    >>>
    >>> x = np.arange(12).reshape([6, 2])
    >>> # create an imbalance dataset
    >>> y = np.zeros(6, np_int_type)
    >>> y[[-1, -2]] = 1
    >>> dataset = TabularDataset.from_xy(x, y, TaskTypes.CLASSIFICATION)
    >>> k_fold = KFold(3, dataset)
    >>> for train_fold, test_fold in k_fold:
    >>>     print(np.vstack([train_fold.dataset.x, test_fold.dataset.x]))
    >>>     print(np.vstack([train_fold.dataset.y, test_fold.dataset.y]))

    """

    def __init__(self,
                 k: int,
                 dataset: TabularDataset,
                 **kwargs):
        if k <= 1:
            raise ValueError("k should be larger than 1 in KFold")
        ratio = 1. / k
        self.n_list = (k - 1) * [ratio]
        self.splitter = DataSplitter(**kwargs).fit(dataset)
        self.split_results = self._order = self._cursor = None

    def __iter__(self):
        self.split_results = self.splitter.split_multiple(self.n_list, return_remained=True)
        self._order = np.random.permutation(len(self.split_results)).tolist()
        self._cursor = 0
        return self

    def __next__(self) -> Tuple[SplitResult, SplitResult]:
        if self._cursor >= len(self._order):
            raise StopIteration
        train_results = self.split_results.copy()
        test_result = train_results.pop(self._order[self._cursor])
        train_result = SplitResult.concat(train_results)
        self._cursor += 1
        return train_result, test_result


class KRandom:
    """
    Util class which can perform k-random data splitting:

    1. X = {x1, x2, ..., xn} -> [X1, X2, ..., Xk]
    2. idx{X1} ≠ idx{X2} ≠ ... ≠ idx{Xk}, where idx{X} = {1, 2, ..., n}
    3. X1 = X2 = ... = Xk = X

    Parameters
    ----------
    k : int, number of folds
    num_test : {int, float}
    * if float and  < 1 : ratio of the test dataset
    * if int   and  > 1 : exact number of test samples
    dataset : TabularDataset, dataset which we want to split
    **kwargs : used to initialize `DataSplitter` instance

    Examples
    ----------
    >>> import numpy as np
    >>>
    >>> from cfdata.types import np_int_type
    >>> from cfdata.tabular.types import TaskTypes
    >>> from cfdata.tabular.wrapper import TabularDataset
    >>> from cfdata.tabular.utils import KRandom
    >>>
    >>> x = np.arange(12).reshape([6, 2])
    >>> # create an imbalance dataset
    >>> y = np.zeros(6, np_int_type)
    >>> y[[-1, -2]] = 1
    >>> dataset = TabularDataset.from_xy(x, y, TaskTypes.CLASSIFICATION)
    >>> k_random = KRandom(3, 2, dataset)
    >>> for train_fold, test_fold in k_random:
    >>>     print(np.vstack([train_fold.dataset.x, test_fold.dataset.x]))
    >>>     print(np.vstack([train_fold.dataset.y, test_fold.dataset.y]))

    """

    def __init__(self,
                 k: int,
                 num_test: Union[int, float],
                 dataset: TabularDataset,
                 **kwargs):
        self._cursor = None
        self.k, self.num_test = k, num_test
        self.splitter = DataSplitter(**kwargs).fit(dataset)

    def __iter__(self):
        self._cursor = 0
        return self

    def __next__(self) -> Tuple[SplitResult, SplitResult]:
        if self._cursor >= self.k:
            raise StopIteration
        self._cursor += 1
        self.splitter.reset()
        test_result, train_result = self.splitter.split_multiple([self.num_test], return_remained=True)
        return train_result, test_result


class ImbalancedSampler(LoggingMixin):
    """
    Util class which can sample imbalance dataset in a balanced way

    Parameters
    ----------
    data : TabularData, data which we want to sample from
    imbalance_threshold : float
    * for binary class cases, if n_pos / n_neg < threshold, we'll treat data as imbalance data
    * for multi  class cases, if n_min_class / n_max_class < threshold, we'll treat data as imbalance data
    shuffle : bool, whether shuffle the returned indices
    sample_method : str, sampling method used in `cftool.misc.Sampler`
    * currently only 'multinomial' is supported
    verbose_level : int, verbose level used in `LoggingMixin`

    Examples
    ----------
    >>> import numpy as np
    >>>
    >>> from cfdata.types import np_int_type
    >>> from cfdata.tabular import TabularData
    >>> from cfdata.tabular.utils import ImbalancedSampler
    >>> from cftool.misc import get_counter_from_arr
    >>>
    >>> n = 20
    >>> x = np.arange(2 * n).reshape([n, 2])
    >>> # create an imbalance dataset
    >>> y = np.zeros([n, 1], np_int_type)
    >>> y[-1] = [1]
    >>> data = TabularData().read(x, y)
    >>> sampler = ImbalancedSampler(data)
    >>> # Counter({1: 12, 0: 8})
    >>> # This may vary, but will be rather balanced
    >>> # You might notice that positive samples are even more than negative samples!
    >>> print(get_counter_from_arr(y[sampler.get_indices()]))

    """

    def __init__(self,
                 data,
                 imbalance_threshold: float = 0.1,
                 *,
                 shuffle: bool = True,
                 sample_method: str = "multinomial",
                 verbose_imbalance: bool = True,
                 verbose_level: int = 2):
        self.data = data
        self.imbalance_threshold = imbalance_threshold
        self._sample_imbalance_flag = True
        self.shuffle, self._n_samples = shuffle, len(data)
        if not self.shuffle or data.task_type is TaskTypes.REGRESSION:
            label_counts = self._label_ratios = self._sampler = None
        else:
            recognizer = data.recognizers[-1]
            label_counter = recognizer.counter
            transform_dict = recognizer.transform_dict
            label_counter = {transform_dict[k]: v for k, v in label_counter.items()}
            label_counts = np.array([label_counter[k] for k in sorted(label_counter)], np_float_type)
            self._label_ratios, max_label_count = label_counts / self._n_samples, label_counts.max()
            if label_counts.min() / max_label_count >= imbalance_threshold:
                self._sampler = None
            else:
                labels = data.processed.y.ravel()
                sample_weights = np.zeros(self._n_samples, np_float_type)
                for i, count in enumerate(label_counts):
                    sample_weights[labels == i] = max_label_count / count
                sample_weights /= sample_weights.sum()
                self._sampler = Sampler(sample_method, sample_weights)

        self._sample_method, self._verbose_level = sample_method, verbose_level
        if verbose_imbalance and self._sampler is not None:
            self.log_msg(
                f"using imbalanced sampler with label counts = {label_counts.tolist()}",
                self.info_prefix, 2
            )

    def __len__(self):
        return self._n_samples

    @property
    def is_imbalance(self) -> bool:
        return self._sampler is not None

    @property
    def sample_imbalance(self) -> bool:
        return self._sample_imbalance_flag

    @property
    def label_ratios(self) -> Union[None, np.ndarray]:
        return self._label_ratios

    def switch_imbalance_status(self, flag: bool) -> None:
        self._sample_imbalance_flag = flag

    def get_indices(self) -> np.ndarray:
        if not self.shuffle or not self._sample_imbalance_flag or not self.is_imbalance:
            indices = np.arange(self._n_samples).astype(np.int64)
        else:
            indices = self._sampler.sample(self._n_samples)
        if self.shuffle:
            np.random.shuffle(indices)
        return indices

    def copy(self) -> "ImbalancedSampler":
        return ImbalancedSampler(
            self.data,
            self.imbalance_threshold,
            shuffle=self.shuffle,
            sample_method=self._sample_method,
            verbose_level=self._verbose_level,
            verbose_imbalance=False
        )


class LabelCollators:
    @staticmethod
    def reg_default(y_batch):
        assert len(y_batch) == 2
        return y_batch[1] - y_batch[0]

    @staticmethod
    def clf_default(y_batch):
        assert len(y_batch) == 2
        return y_batch[1] == y_batch[0]


class DataLoader:
    """
    Util class which can generated batches from `ImbalancedSampler`

    Examples
    ----------
    >>> import numpy as np
    >>>
    >>> from cfdata.types import np_int_type
    >>> from cfdata.tabular import TabularData
    >>> from cfdata.tabular.utils import DataLoader
    >>> from cfdata.tabular.utils import ImbalancedSampler
    >>> from cftool.misc import get_counter_from_arr
    >>>
    >>> n = 20
    >>> x = np.arange(2 * n).reshape([n, 2])
    >>> y = np.zeros([n, 1], np_int_type)
    >>> y[-1] = [1]
    >>> data = TabularData().read(x, y)
    >>> sampler = ImbalancedSampler(data)
    >>> loader = DataLoader(16, sampler)
    >>> y_batches = []
    >>> for x_batch, y_batch in loader:
    >>>     y_batches.append(y_batch)
    >>>     # (16, 1) (16, 1)
    >>>     # (4, 1) (4, 1)
    >>>     print(x_batch.shape, y_batch.shape)
    >>> # Counter({1: 11, 0: 9})
    >>> print(get_counter_from_arr(np.vstack(y_batches).ravel()))

    """

    def __init__(self,
                 batch_size: int,
                 sampler: ImbalancedSampler,
                 *,
                 n_siamese: int = 1,
                 label_collator: callable = None,
                 verbose_level: int = 2):
        self._indices_in_use = None
        self._verbose_level = verbose_level
        self.data = sampler.data
        self._n_siamese, self._label_collator = n_siamese, label_collator
        self._n_samples, self.sampler = len(self.data), sampler
        self.batch_size = min(self._n_samples, batch_size)
        self._verbose_level = verbose_level

    def __len__(self):
        n_iter = int(self._n_samples / self.batch_size)
        return n_iter + int(n_iter * self.batch_size < self._n_samples)

    def __iter__(self):
        self._reset()
        return self

    def __next__(self):
        data_next = self._get_next_batch()
        if self._n_siamese == 1:
            return data_next
        all_data = [data_next] if self._check_full_batch(data_next) else []
        while len(all_data) < self._n_siamese:
            data_next = self._get_next_batch()
            if self._check_full_batch(data_next):
                all_data.append(data_next)
        x_batch, y_batch = zip(*all_data)
        if self._label_collator is not None:
            y_batch = self._label_collator(y_batch)
        return x_batch, y_batch

    @property
    def enabled_sampling(self) -> bool:
        return self.sampler.sample_imbalance

    @enabled_sampling.setter
    def enabled_sampling(self, value: bool):
        self.sampler.switch_imbalance_status(value)

    def _reset(self):
        reset_caches = {
            "_indices_in_use": self.sampler.get_indices(),
            "_siamese_cursor": 0, "_cursor": -1
        }
        for attr, init_value in reset_caches.items():
            setattr(self, attr, init_value)

    def _get_next_batch(self):
        n_iter, self._cursor = len(self), self._cursor + 1
        if self._cursor == n_iter * self._n_siamese:
            raise StopIteration
        if self._n_siamese > 1:
            new_siamese_cursor = int(self._cursor / n_iter)
            if new_siamese_cursor > self._siamese_cursor:
                self._siamese_cursor = new_siamese_cursor
                self._indices_in_use = self.sampler.get_indices()
        start = (self._cursor - n_iter * self._siamese_cursor) * self.batch_size
        end = start + self.batch_size
        batch = self.data[self._indices_in_use[start:end]]
        return batch

    def _check_full_batch(self, batch):
        if len(batch[0]) == self.batch_size:
            return True
        return False

    def copy(self) -> "DataLoader":
        return DataLoader(
            self.batch_size,
            self.sampler.copy(),
            n_siamese=self._n_siamese,
            label_collator=self._label_collator,
            verbose_level=self._verbose_level
        )


__all__ = [
    "SplitResult", "DataSplitter", "KFold", "KRandom",
    "ImbalancedSampler", "LabelCollators", "DataLoader"
]



import numpy as np

from typing import *
from abc import ABC, abstractmethod

from ..types import *
from ..recognizer import Recognizer
from cftool.misc import register_core

converter_dict: Dict[str, Type["Converter"]] = {}


class Converter(ABC):
    def __init__(self,
                 recognizer: Recognizer,
                 *,
                 inplace: bool = False,
                 **kwargs):
        self._inplace = inplace
        self._recognizer = recognizer
        self._initialize(**kwargs)
        self._fit()

    def __str__(self):
        return f"Converter({self.info.column_type})"

    __repr__ = __str__

    @abstractmethod
    def _fit(self) -> "Converter":
        pass

    @abstractmethod
    def _convert(self,
                 flat_arr: flat_arr_type) -> np.ndarray:
        pass

    @abstractmethod
    def _recover(self,
                 flat_arr: flat_arr_type) -> np.ndarray:
        pass

    @property
    def info(self):
        return self._recognizer.info

    @property
    def converted_input(self):
        if getattr(self, "_converted_features", None) is None:
            self._converted_features = self.convert(self.info.flat_arr)
        return self._converted_features

    def _initialize(self, **kwargs) -> None:
        pass

    def convert(self,
                flat_arr: flat_arr_type) -> np.ndarray:
        if not self._inplace:
            flat_arr = flat_arr.copy()
        return self._convert(flat_arr)

    def recover(self,
                flat_arr: flat_arr_type,
                *,
                inplace: bool = False) -> np.ndarray:
        if not inplace:
            flat_arr = flat_arr.copy()
        return self._recover(flat_arr)

    @classmethod
    def make_with(cls,
                  recognizer: Recognizer,
                  *,
                  inplace: bool = False,
                  **kwargs) -> "Converter":
        key = recognizer.info.column_type.value
        return converter_dict[key](recognizer, inplace=inplace, **kwargs)

    @classmethod
    def register(cls, name):
        global converter_dict
        return register_core(name, converter_dict)


__all__ = ["Converter", "converter_dict"]



import os
import copy
import logging

import numpy as np

from typing import *
from cftool.misc import timing_context, SavingMixin

from .types import *
from .recognizer import *
from .converters import *
from .processors import *
from .utils import DataSplitter
from ..base import DataBase
from ..types import np_int_type


# TODO : Add outlier detection
class TabularData(DataBase):
    def __init__(self,
                 *,
                 task_type: TaskTypes = None,
                 label_name: str = None,
                 string_label: Union[bool, None] = None,
                 numerical_label: Union[bool, None] = None,
                 categorical_label: Union[bool, None] = None,
                 column_names: Dict[int, str] = None,
                 valid_columns: List[int] = None,
                 string_columns: List[int] = None,
                 numerical_columns: List[int] = None,
                 categorical_columns: List[int] = None,
                 process_methods: Union[str, None, Dict[int, str]] = "auto",
                 default_numerical_process: str = "normalize",
                 default_categorical_process: str = "one_hot",
                 label_process_method: str = None,
                 numerical_threshold: float = None,
                 trigger_logging: bool = False,
                 verbose_level: int = 1):
        if task_type is not None:
            if task_type is TaskTypes.CLASSIFICATION:
                if numerical_label:
                    raise ValueError("numerical labels are invalid in CLASSIFICATION tasks")
            else:
                if string_label:
                    raise ValueError("string labels are invalid in REGRESSION tasks")
                if categorical_label:
                    raise ValueError("categorical labels are invalid in REGRESSION tasks")
        self._task_type = task_type
        self.label_name = label_name
        self.string_label = string_label
        self.numerical_label = numerical_label
        self.categorical_label = categorical_label
        self._column_names = column_names
        self._valid_columns = valid_columns
        self._string_columns = string_columns
        self._numerical_columns = numerical_columns
        self._categorical_columns = categorical_columns
        self._process_methods = process_methods
        self._default_numerical_process = default_numerical_process
        self._default_categorical_process = default_categorical_process
        self._label_process_method = label_process_method
        self._numerical_threshold = numerical_threshold
        self._is_file = self._is_arr = False
        self._num_classes = None
        self._label_idx = self._has_column_names = None
        self._delim = self._quote_char = None
        self._raw = self._converted = self._processed = None
        self._recognizers = self._converters = self._processors = None
        self._verbose_level = verbose_level
        self._init_logging(verbose_level, trigger=trigger_logging)
        self.excludes = set()

    def __len__(self):
        return len(self._processed.x)

    def __getitem__(self, indices: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        x, y = self.processed.xy
        y_batch = None if y is None else y[indices]
        return x[indices], y_batch

    def __eq__(self, other: "TabularData"):
        if self.raw != other.raw:
            return False
        if self.converted != other.converted:
            return False
        return self.processed == other.processed

    @property
    def cache_excludes(self):
        return {"_recognizers", "_converters", "_processors"}

    @property
    def data_tuple_base(self) -> Type[NamedTuple]:
        return DataTuple

    @property
    def data_tuple_attributes(self) -> List[str]:
        return ["_raw", "_converted", "_processed"]

    @property
    def raw(self) -> DataTuple:
        return self._raw

    @property
    def converted(self) -> DataTuple:
        return self._converted

    @property
    def processed(self) -> DataTuple:
        return self._processed

    @property
    def recognizers(self) -> Dict[int, Recognizer]:
        return self._recognizers

    @property
    def converters(self) -> Dict[int, Converter]:
        return self._converters

    @property
    def processors(self) -> Dict[int, Processor]:
        return self._processors

    @property
    def raw_dim(self) -> int:
        return len(self._raw.x[0])

    @property
    def processed_dim(self) -> int:
        return self._processed.x.shape[1]

    @property
    def task_type(self) -> Union[TaskTypes, None]:
        if self._task_type is not None:
            return self._task_type
        if self._recognizers[-1] is None:
            return
        self._task_type = TaskTypes.from_column_type(self._recognizers[-1].info.column_type)
        return self._task_type

    @property
    def column_names(self) -> Dict[int, str]:
        if self._column_names is None:
            self._column_names = {}
        for i in range(self.raw_dim):
            self._column_names.setdefault(i, str(i))
        return self._column_names

    def _get_prior_dict(self, attr: str) -> Dict[int, Union[bool, None]]:
        prior_dict_attr = f"{attr}_dict"
        prior_dict = getattr(self, prior_dict_attr, None)
        if prior_dict is None:
            prior_columns = getattr(self, attr, None)
            prior_dict = {
                i: None if prior_columns is None else i in prior_columns
                for i in range(self.raw_dim)
            }
            setattr(self, prior_dict_attr, prior_dict)
        return prior_dict

    @property
    def prior_valid_columns(self) -> Dict[int, Union[bool, None]]:
        return self._get_prior_dict("_valid_columns")

    @property
    def prior_string_columns(self) -> Dict[int, Union[bool, None]]:
        return self._get_prior_dict("_string_columns")

    @property
    def prior_numerical_columns(self) -> Dict[int, Union[bool, None]]:
        return self._get_prior_dict("_numerical_columns")

    @property
    def prior_categorical_columns(self) -> Dict[int, Union[bool, None]]:
        return self._get_prior_dict("_categorical_columns")

    @property
    def is_clf(self) -> bool:
        return self.task_type is TaskTypes.CLASSIFICATION

    @property
    def is_reg(self) -> bool:
        return self.task_type is TaskTypes.REGRESSION

    @property
    def is_file(self) -> bool:
        return self._is_file

    @property
    def num_classes(self) -> int:
        if self.is_reg:
            return 0
        if self._num_classes is None:
            self._num_classes = self._processed.y.max().item() + 1
        return self._num_classes

    # Core

    @staticmethod
    def _flatten(data: data_type) -> data_type:
        if isinstance(data, list):
            return sum(data, [])
        return data.ravel()

    def _core_fit(self) -> "TabularData":
        with timing_context(self, "convert"):
            # convert features
            features = self._raw.xT
            converted_features = []
            self._recognizers, self._converters = {}, {}
            for i, flat_arr in enumerate(features):
                column_name = self.column_names[i]
                is_valid = self.prior_valid_columns[i]
                is_string = self.prior_string_columns[i]
                is_numerical = self.prior_numerical_columns[i]
                is_categorical = self.prior_categorical_columns[i]
                if i == self.raw_dim - 1 == len(self.excludes):
                    if i > 0:
                        self.log_msg(
                            f"last column {column_name} is forced to be valid "
                            "because previous columns are all excluded", self.warning_prefix,
                            verbose_level=2, msg_level=logging.WARNING
                        )
                    is_valid = True
                kwargs = {
                    "is_valid": is_valid,
                    "is_string": is_string,
                    "is_numerical": is_numerical,
                    "is_categorical": is_categorical,
                }
                if self._numerical_threshold is not None:
                    kwargs["numerical_threshold"] = self._numerical_threshold
                with timing_context(self, "fit recognizer"):
                    recognizer = self._recognizers[i] = Recognizer(column_name, **kwargs).fit(flat_arr)
                if not recognizer.info.is_valid:
                    self.log_msg(recognizer.info.msg, self.warning_prefix, 2, logging.WARNING)
                    self.excludes.add(i)
                    continue
                with timing_context(self, "fit converter"):
                    converter = self._converters[i] = Converter.make_with(recognizer)
                converted_features.append(converter.converted_input)
            # convert labels
            if self._raw.y is None:
                converted_labels = self._recognizers[-1] = self._converters[-1] = None
            else:
                with timing_context(self, "fit recognizer"):
                    recognizer = self._recognizers[-1] = Recognizer(
                        self.label_name,
                        is_label=True,
                        task_type=self._task_type,
                        is_valid=True,
                        is_string=self.string_label,
                        is_numerical=self.numerical_label,
                        is_categorical=self.categorical_label,
                        numerical_threshold=1.
                    ).fit(self._flatten(self._raw.y))
                with timing_context(self, "fit converter"):
                    converter = self._converters[-1] = Converter.make_with(recognizer)
                converted_labels = converter.converted_input.reshape([-1, 1])
        converted_x = np.vstack(converted_features).T
        with timing_context(self, "process"):
            # process features
            self._processors = {}
            processed_features = []
            previous_processors = []
            idx = 0
            while idx < self.raw_dim:
                if idx in self.excludes:
                    idx += 1
                    continue
                column_type = self._converters[idx].info.column_type
                if self._process_methods is None:
                    method = None
                elif isinstance(self._process_methods, str):
                    method = self._process_methods
                else:
                    method = self._process_methods.get(idx, "auto")
                if method is None:
                    method = "identical"
                elif method == "auto":
                    if column_type is ColumnTypes.NUMERICAL:
                        method = self._default_numerical_process
                    else:
                        method = self._default_categorical_process
                processor = self._processors[idx] = processor_dict[method](previous_processors.copy())
                previous_processors.append(processor)
                columns = converted_x[..., processor.input_indices]
                with timing_context(self, "fit processor"):
                    processor.fit(columns)
                with timing_context(self, "process with processor"):
                    processed_features.append(processor.process(columns))
                idx += processor.input_dim
            # process labels
            if converted_labels is None:
                processed_labels = self._processors[-1] = None
            else:
                column_type = self._converters[-1].info.column_type
                method = None
                if self._label_process_method is not None:
                    method = self._label_process_method
                if method is None:
                    method = "normalize" if column_type is ColumnTypes.NUMERICAL else "identical"
                with timing_context(self, "fit processor"):
                    processor = self._processors[-1] = processor_dict[method]([]).fit(converted_labels)
                with timing_context(self, "process with processor"):
                    processed_labels = processor.process(converted_labels)
        if self.task_type is TaskTypes.CLASSIFICATION and converted_labels is not None:
            converted_labels = converted_labels.astype(np_int_type)
            processed_labels = processed_labels.astype(np_int_type)
        self._converted = DataTuple(converted_x, converted_labels)
        self._processed = DataTuple(np.hstack(processed_features), processed_labels)
        self._valid_columns = [col for col in range(self.raw_dim) if col not in self.excludes]
        self._valid_columns_dict = None
        return self

    def _read_from_file(self,
                        file_path: str,
                        *,
                        label_idx: int = None,
                        has_column_names: bool = None,
                        quote_char: str = None,
                        delim: str = None) -> "TabularData":
        self._is_file = True
        self._label_idx, self._has_column_names = label_idx, has_column_names
        self._delim, self._quote_char = delim, quote_char
        with timing_context(self, "_read_file"):
            x, y = self.read_file(file_path)
        self._raw = DataTuple.with_transpose(x, y)
        return self._core_fit()

    def _read_from_arr(self,
                       x: data_type,
                       y: data_type) -> "TabularData":
        self._is_arr = True
        self._raw = DataTuple.with_transpose(x, y)
        return self._core_fit()

    def _transform_labels(self,
                          raw: DataTuple) -> Tuple[np.ndarray, np.ndarray]:
        if raw.y is None:
            converted_labels = transformed_labels = None
        else:
            converted_labels = self._converters[-1].convert(self._flatten(raw.y))
            transformed_labels = self._processors[-1].process(converted_labels.reshape([-1, 1]))
        if self.task_type is TaskTypes.CLASSIFICATION:
            if converted_labels is not None:
                converted_labels = converted_labels.astype(np_int_type)
                transformed_labels = transformed_labels.astype(np_int_type)
        return converted_labels, transformed_labels

    def _transform(self,
                   raw: DataTuple,
                   return_converted: bool) -> Union[DataTuple, Tuple[DataTuple, DataTuple]]:
        # transform features
        features = raw.xT
        converted_features = np.vstack([
            self._converters[i].convert(flat_arr)
            for i, flat_arr in enumerate(features) if i not in self.excludes
        ])
        idx = 0
        processed = []
        while idx < self.raw_dim:
            if idx in self.excludes:
                idx += 1
                continue
            processor = self._processors[idx]
            input_indices = processor.input_indices
            columns = processor.process(converted_features[input_indices].T)
            processed.append(columns)
            idx += processor.input_dim
        transformed_features = np.hstack(processed)
        # transform labels
        converted_labels, transformed_labels = self._transform_labels(raw)
        # aggregate
        transformed = DataTuple(transformed_features, transformed_labels)
        if not return_converted:
            return transformed
        converted = DataTuple(converted_features.T, converted_labels)
        return converted, transformed

    def _get_raw(self,
                 x: Union[str, data_type],
                 y: data_type = None,
                 *,
                 contains_labels: bool = True) -> DataTuple:
        if self._is_file:
            if isinstance(x, str):
                if y is not None:
                    raise ValueError(f"x refers to file_path but y is still provided ({y}), which is illegal")
                x, y = self.read_file(x, contains_labels=contains_labels)
        return DataTuple.with_transpose(x, y)

    # API

    def read_file(self,
                  file_path: str,
                  *,
                  contains_labels: bool = True) -> Tuple[raw_data_type, raw_data_type]:
        ext = os.path.splitext(file_path)[1][1:]
        set_default = lambda n, default: n if n is not None else default
        if ext == "txt":
            has_column_names, delim, quote_char = map(
                set_default,
                [self._has_column_names, self._delim, self._quote_char],
                [False, " ", None]
            )
        elif ext == "csv":
            has_column_names, delim, quote_char = map(
                set_default,
                [self._has_column_names, self._delim, self._quote_char],
                [True, ",", '"']
            )
        else:
            raise NotImplementedError(f"file type '{ext}' not recognized")
        with open(file_path, "r") as f:
            first_row = None
            if has_column_names:
                first_row = column_names = f.readline().strip().split(delim)
                self._column_names = {i: name for i, name in enumerate(column_names)}
            data = []
            for line in f:
                line = ["nan" if not elem else elem for elem in line.strip().split(delim)]
                if quote_char is not None:
                    startswith_quote = [elem.startswith(quote_char) for elem in line]
                    endswith_quote = [elem.endswith(quote_char) for elem in line]
                    merge_start, merge_intervals = None, []
                    for i, (startswith, endswith) in enumerate(zip(startswith_quote, endswith_quote)):
                        if startswith and not endswith:
                            merge_start = i
                            continue
                        if endswith and not startswith and merge_start is not None:
                            merge_intervals.append((merge_start, i + 1))
                            merge_start = None
                            continue
                    idx, new_line = 0, []
                    for start, end in merge_intervals:
                        if start > idx:
                            new_line += line[idx:start]
                        new_line.append(delim.join(line[start:end]))
                        idx = end
                    if idx < len(line):
                        new_line += line[idx:len(line)]
                    line = new_line
                if first_row is None:
                    first_row = line
                else:
                    assert len(first_row) == len(line), "num_features are not identical"
                data.append(line)
        if not contains_labels:
            if len(data[0]) == len(self._raw.x[0]) + 1:
                raise ValueError("file contains labels but 'contains_labels=False' passed in")
            return data, None

        if self._column_names is None or self.label_name is None:
            if self._label_idx is None:
                self._label_idx = -1
        else:
            reverse_column_names = dict(map(reversed, self._column_names.items()))
            self._label_idx = reverse_column_names.get(self.label_name)
            if self._label_idx is None:
                raise ValueError(
                    f"'{self.label_name}' is not included in column names "
                    f"({[self._column_names[i] for i in range(len(self._column_names))]})"
                )
        if self._label_idx < 0:
            self._label_idx = len(data[0]) + self._label_idx
        x = [line[:self._label_idx] + line[self._label_idx+1:] for line in data]
        y = [line[self._label_idx:self._label_idx+1] for line in data]
        return x, y

    def read(self,
             x: Union[str, data_type],
             y: Union[int, data_type] = None,
             **kwargs) -> "TabularData":
        if isinstance(x, str):
            self._read_from_file(x, label_idx=y, **kwargs)
        else:
            if isinstance(y, int):
                y = None
            self._read_from_arr(x, y)
        self.log_timing()
        return self

    def split(self,
              n: Union[int, float]) -> Tuple["TabularData", "TabularData"]:
        splitter = DataSplitter(shuffle=False).fit(self.to_dataset())
        split = splitter.split(n)
        split_indices = split.corresponding_indices
        remained_indices = split.remaining_indices
        raw, converted, processed = self._raw, self._converted, self._processed
        p1, p2 = map(copy.copy, [self] * 2)
        p1._raw, p1._converted, p1._processed = map(
            DataTuple.split_with,
            [raw, converted, processed], [split_indices] * 3
        )
        p2._raw, p2._converted, p2._processed = map(
            DataTuple.split_with,
            [raw, converted, processed], [remained_indices] * 3
        )
        return p1, p2

    def copy_to(self,
                x: Union[str, data_type],
                y: data_type = None,
                *,
                contains_labels: bool = True) -> "TabularData":
        copied = copy.copy(self)
        raw = copied._raw = self._get_raw(x, y, contains_labels=contains_labels)
        copied._converted, copied._processed = self._transform(raw, True)
        return copied

    def transform(self,
                  x: Union[str, data_type],
                  y: data_type = None,
                  *,
                  return_converted: bool = False) -> Union[DataTuple, Tuple[DataTuple, DataTuple]]:
        raw = self._get_raw(x, y)
        return self._transform(raw, return_converted)

    def transform_labels(self,
                         y: data_type,
                         *,
                         return_converted: bool = False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        raw = DataTuple(None, y)
        converted_labels, transformed_labels = self._transform_labels(raw)
        if not return_converted:
            return transformed_labels
        return converted_labels, transformed_labels

    def recover_labels(self,
                       y: np.ndarray,
                       *,
                       inplace: bool = False) -> np.ndarray:
        process_recovered = self._processors[-1].recover(y, inplace=inplace)
        convert_recovered = self.converters[-1].recover(process_recovered.ravel(), inplace=inplace)
        return convert_recovered.reshape([-1, 1])

    @classmethod
    def load(cls,
             folder: str,
             *,
             compress: bool = True,
             verbose_level: int = 0) -> "TabularData":
        data = cls(verbose_level=verbose_level)
        SavingMixin.load(data, folder, compress=compress)
        is_file, is_arr = data._is_file, data._is_arr
        data.read(*data._raw[:2])
        data._is_file, data._is_arr = is_file, is_arr
        return data

    def to_dataset(self) -> TabularDataset:
        return TabularDataset(*self.processed.xy, task_type=self.task_type)

    @classmethod
    def from_dataset(cls,
                     dataset: TabularDataset,
                     **kwargs):
        task_type = kwargs.pop("task_type", dataset.task_type)
        return cls(task_type=task_type, **kwargs).read(*dataset.xy)


__all__ = ["TabularData"]



cimport cython
cimport numpy as np

import unicodedata
import numpy as np
from libc.math cimport isnan
from collections import defaultdict


ctypedef fused arr:
    list
    tuple

# inner-core functions

def _is_numeric(s):
    try:
        s = float(s)
        return True
    except (TypeError, ValueError):
        try:
            unicodedata.numeric(s)
            return True
        except (TypeError, ValueError):
            return False


# api

@cython.wraparound(False)
@cython.boundscheck(False)
def transform_flat_data_with_dict(np.ndarray[np.float32_t, ndim=1] flat_data, transform_dict):
    cdef dict d = transform_dict
    cdef float elem
    cdef unsigned int i, n=len(flat_data)
    for i in range(n):
        elem = float(flat_data[i])
        if isnan(elem):
            flat_data[i] = d.get("nan", 0)
        else:
            flat_data[i] = d.get(elem, 0)
    return flat_data


@cython.wraparound(False)
@cython.boundscheck(False)
def is_all_numeric(arr x):
    cdef unsigned int i, n=len(x)
    for i in range(n):
        if not _is_numeric(x[i]):
            return False
    return True


@cython.wraparound(False)
@cython.boundscheck(False)
def flat_arr_to_float32(arr x):
    cdef unsigned i, n=len(x)
    cdef np.ndarray[np.float32_t, ndim=1] rs=np.zeros(n, dtype=np.float32)
    for i in range(n):
        rs[i] = float(x[i])
    return rs



import math
import numpy as np

from cftool.misc import is_numeric


def naive_transform_flat_data_with_dict(flat_data, transform_dict):
    for i, elem in enumerate(flat_data):
        elem = float(elem)
        if math.isnan(elem):
            flat_data[i] = transform_dict.get("nan", 0)
        else:
            flat_data[i] = transform_dict.get(elem, 0)
    return flat_data


def naive_is_all_numeric(arr):
    return all(map(is_numeric, arr))


def naive_flat_arr_to_float32(arr):
    return np.asarray(arr, np.float32)


__all__ = [
    "naive_transform_flat_data_with_dict",
    "naive_is_all_numeric", "naive_flat_arr_to_float32"
]



try:
    from .cython_utils import *
except ImportError:
    raise


def _to_list(arr):
    if isinstance(arr, np.ndarray):
        arr = arr.tolist()
    return arr


def c_transform_flat_data_with_dict(flat_data, transform_dict):
    return transform_flat_data_with_dict(flat_data, transform_dict)


def c_is_all_numeric(arr):
    return is_all_numeric(_to_list(arr))


def c_flat_arr_to_float32(arr):
    return flat_arr_to_float32(_to_list(arr))


__all__ = [
    "c_transform_flat_data_with_dict",
    "c_is_all_numeric", "c_flat_arr_to_float32"
]



from .base import *
from .normalize import Normalize
from .min_max import MinMax
from .function import Lambda
from .one_hot import OneHot
from .identical import Identical


__all__ = ["Processor", "processor_dict"]



import numpy as np

from ..base import Processor


@Processor.register("identical")
class Identical(Processor):
    @property
    def input_dim(self) -> int:
        return 1

    @property
    def output_dim(self) -> int:
        return 1

    def fit(self,
            columns: np.ndarray) -> Processor:
        return self

    def _process(self,
                 columns: np.ndarray) -> np.ndarray:
        return columns

    def _recover(self,
                 processed_columns: np.ndarray) -> np.ndarray:
        return processed_columns


__all__ = ["Identical"]



import numpy as np

from ...types import *
from ....types import *
from ..base import Converter
from ....misc.c import transform_flat_data_with_dict


@Converter.register("categorical")
class CategoricalConverter(Converter):
    def _fit(self) -> "Converter":
        assert self.info.is_valid and self.info.is_categorical
        self._transform_dict = self._recognizer.transform_dict
        self._reverse_transform_dict = dict(map(reversed, self._transform_dict.items()))
        return self

    def _convert(self,
                 flat_arr: flat_arr_type) -> np.ndarray:
        flat_arr = np.asarray(flat_arr, np_float_type)
        if not self.info.need_transform:
            return flat_arr
        return transform_flat_data_with_dict(flat_arr, self._transform_dict)

    def _recover(self,
                 flat_arr: flat_arr_type) -> np.ndarray:
        flat_arr = np.asarray(flat_arr, np_float_type)
        if not self.info.need_transform:
            return flat_arr
        return transform_flat_data_with_dict(flat_arr, self._reverse_transform_dict)


__all__ = ["CategoricalConverter"]



try:
    from .cython_wrappers import c_transform_flat_data_with_dict as transform_flat_data_with_dict
    from .cython_wrappers import c_flat_arr_to_float32 as flat_arr_to_float32
    from .cython_wrappers import c_is_all_numeric as is_all_numeric
except ImportError:
    from .cython_substitute import naive_transform_flat_data_with_dict as transform_flat_data_with_dict
    from .cython_substitute import naive_flat_arr_to_float32 as flat_arr_to_float32
    from .cython_substitute import naive_is_all_numeric as is_all_numeric


__all__ = [
    "transform_flat_data_with_dict",
    "flat_arr_to_float32", "is_all_numeric"
]



import numpy as np

from typing import *
from abc import ABC, abstractmethod
from cftool.misc import register_core

processor_dict: Dict[str, Type["Processor"]] = {}


class Processor(ABC):
    def __init__(self,
                 previous_processors: List["Processor"],
                 *,
                 inplace: bool = False,
                 **kwargs):
        self._caches = {}
        self._inplace = inplace
        self._previous_processors = previous_processors
        start_idx = sum([processor.input_dim for processor in self._previous_processors])
        self._col_indices = [start_idx + i for i in range(self.input_dim)]
        self._initialize(**kwargs)

    def __str__(self):
        return f"{type(self).__name__}()"

    __repr__ = __str__

    @property
    @abstractmethod
    def input_dim(self) -> int:
        pass

    @property
    @abstractmethod
    def output_dim(self) -> int:
        pass

    @abstractmethod
    def fit(self,
            columns: np.ndarray) -> "Processor":
        pass

    @abstractmethod
    def _process(self,
                 columns: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def _recover(self,
                 processed_columns: np.ndarray) -> np.ndarray:
        pass

    @property
    def input_indices(self) -> List[int]:
        return self._col_indices

    @property
    def output_indices(self) -> List[int]:
        previous_dimensions = sum([method.output_dim for method in self._previous_processors])
        return list(range(previous_dimensions, previous_dimensions + self.output_dim))

    def _initialize(self, **kwargs) -> None:
        pass

    def process(self,
                columns: np.ndarray) -> np.ndarray:
        if not self._inplace:
            columns = columns.copy()
        return self._process(columns)

    def recover(self,
                columns: np.ndarray,
                *,
                inplace: bool = False) -> np.ndarray:
        if not inplace:
            columns = columns.copy()
        return self._recover(columns)

    @classmethod
    def register(cls, name):
        global processor_dict
        return register_core(name, processor_dict)


__all__ = ["Processor", "processor_dict"]



import math
import numpy as np

from typing import *
from enum import Enum
from functools import partial
from sklearn.utils import Bunch
from sklearn.datasets import *

from ..types import *

flat_arr_type = Union[list, np.ndarray]
raw_data_type = Union[List[List[Union[str, float]]], None]
data_type = Union[raw_data_type, np.ndarray]


class DataTuple(NamedTuple):
    x: data_type
    y: data_type
    xT: data_type = None

    def __eq__(self, other: "DataTuple"):
        self_x_is_list = isinstance(self.x, list)
        other_x_is_list = isinstance(other.x, list)
        if self_x_is_list and not other_x_is_list:
            return False
        if not self_x_is_list and other_x_is_list:
            return False
        if self_x_is_list:
            x_allclose = self.x == other.x
        else:
            if isinstance(self.x[0][0], np.str_):
                x_allclose = self.x.tolist() == other.x.tolist()
            else:
                x_allclose = np.allclose(self.x, other.x, equal_nan=True)
        if not x_allclose:
            return False
        if self.y is None and other.y is not None:
            return False
        if self.y is not None and other.y is None:
            return False
        if self.y is None and other.y is None:
            return True
        self_y_is_list = isinstance(self.y, list)
        other_y_is_list = isinstance(other.y, list)
        if self_y_is_list and not other_y_is_list:
            return False
        if not self_y_is_list and other_y_is_list:
            return False
        if self_y_is_list:
            return self.y == other.y
        if isinstance(self.y[0][0], np.str_):
            return self.y.tolist() == other.y.tolist()
        return np.allclose(self.y, other.y, equal_nan=True)

    def __ne__(self, other: "DataTuple"):
        return not self == other

    @property
    def xy(self) -> Tuple[data_type, data_type]:
        return self.x, self.y

    def split_with(self,
                   indices: Union[np.ndarray, List[int]]) -> "DataTuple":
        x = y = xt = None
        if isinstance(self.x, np.ndarray) or isinstance(self.y, np.ndarray):
            if self.x is not None:
                x = self.x[indices]
            if self.y is not None:
                y = self.y[indices]
            if self.xT is not None:
                xt = self.xT[..., indices]
        else:
            if self.x is not None:
                x = [self.x[i] for i in indices]
            if self.y is not None:
                y = [self.y[i] for i in indices]
            if self.xT is not None:
                xt = [[line[i] for i in indices] for line in self.xT]
        return DataTuple(x, y, xt)

    @classmethod
    def with_transpose(cls,
                       x: data_type,
                       y: data_type):
        xt = x.T if not isinstance(x, list) else list(map(list, zip(*x)))
        return DataTuple(x, y, xt)


class ColumnTypes(Enum):
    NUMERICAL = "numerical"
    CATEGORICAL = "categorical"
    STRING = "string"


class TaskTypes(Enum):
    REGRESSION = "reg"
    CLASSIFICATION = "clf"

    @classmethod
    def from_str(cls,
                 task_type: str) -> "TaskTypes":
        if task_type == "reg":
            return cls.REGRESSION
        if task_type == "clf":
            return cls.CLASSIFICATION
        raise ValueError(f"task_type '{task_type}' is not recognized")

    @classmethod
    def from_column_type(cls,
                         column_type: ColumnTypes) -> "TaskTypes":
        if column_type is ColumnTypes.NUMERICAL:
            return cls.REGRESSION
        return cls.CLASSIFICATION


class FeatureInfo(NamedTuple):
    contains_nan: Union[bool, None]
    flat_arr: Union[flat_arr_type, None]
    is_valid: bool = True
    nan_mask: np.ndarray = None
    need_transform: bool = None
    column_type: ColumnTypes = ColumnTypes.NUMERICAL
    unique_values_sorted_by_counts: np.ndarray = None
    msg: str = None

    @property
    def is_string(self) -> bool:
        return self.column_type is ColumnTypes.STRING

    @property
    def is_categorical(self) -> bool:
        return self.column_type is ColumnTypes.CATEGORICAL

    @property
    def is_numerical(self) -> bool:
        return self.column_type is ColumnTypes.NUMERICAL


class TabularDataset(NamedTuple):
    x: np.ndarray
    y: np.ndarray
    task_type: Union[None, TaskTypes] = None
    label_name: Union[None, str] = "label"
    label_names: Union[None, List[str]] = None
    feature_names: Union[None, List[str]] = None

    def __len__(self):
        return self.x.shape[0]

    @property
    def xy(self) -> Tuple[np.ndarray, np.ndarray]:
        return self.x, self.y

    @property
    def is_clf(self) -> bool:
        return self.task_type is TaskTypes.CLASSIFICATION

    @property
    def is_reg(self) -> bool:
        return self.task_type is TaskTypes.REGRESSION

    @property
    def num_features(self) -> int:
        return self.x.shape[1]

    @property
    def num_classes(self) -> int:
        if self.task_type == "reg":
            return 0
        return self.y.max().item() + 1

    @staticmethod
    def to_task_type(x: np.ndarray,
                     y: np.ndarray,
                     task_type: TaskTypes) -> Tuple[np.ndarray, np.ndarray]:
        x = x.astype(np_float_type)
        is_clf = task_type is TaskTypes.CLASSIFICATION
        y = y.reshape([-1, 1]).astype(np_int_type if is_clf else np_float_type)
        return x, y

    @classmethod
    def from_bunch(cls,
                   bunch: Bunch,
                   task_type: TaskTypes) -> "TabularDataset":
        x, y = TabularDataset.to_task_type(bunch.data, bunch.target, task_type)
        label_names = bunch.get("target_names")
        feature_names = bunch.get("feature_names")
        return TabularDataset(x, y, task_type, label_names=label_names, feature_names=feature_names)

    @classmethod
    def from_xy(cls,
                x: np.ndarray,
                y: np.ndarray,
                task_type: TaskTypes) -> "TabularDataset":
        x, y = cls.to_task_type(x, y, task_type)
        return TabularDataset(x, y, task_type)

    # scikit-learn datasets

    @classmethod
    def iris(cls) -> "TabularDataset":
        return cls.from_bunch(load_iris(), TaskTypes.CLASSIFICATION)

    @classmethod
    def boston(cls) -> "TabularDataset":
        return cls.from_bunch(load_boston(), TaskTypes.REGRESSION)

    @classmethod
    def digits(cls) -> "TabularDataset":
        return cls.from_bunch(load_digits(), TaskTypes.CLASSIFICATION)

    @classmethod
    def breast_cancer(cls) -> "TabularDataset":
        return cls.from_bunch(load_breast_cancer(), TaskTypes.CLASSIFICATION)

    # artificial datasets

    @classmethod
    def xor(cls, *,
            size: int = 100,
            scale: float = 1.) -> "TabularDataset":
        x = np.random.randn(size) * scale
        y = np.random.randn(size) * scale
        z = (x * y >= 0).astype(np_int_type)
        return TabularDataset.from_xy(np.c_[x, y].astype(np_float_type), z, TaskTypes.CLASSIFICATION)

    @classmethod
    def spiral(cls, *,
               size: int = 50,
               scale: float = 4.,
               nun_spirals: int = 7,
               num_classes: int = 7) -> "TabularDataset":
        xs = np.zeros((size * nun_spirals, 2), dtype=np_float_type)
        ys = np.zeros(size * nun_spirals, dtype=np_int_type)
        pi = math.pi
        for i in range(nun_spirals):
            ix = range(size * i, size * (i + 1))
            r = np.linspace(0.0, 1, size + 1)[1:]
            t_start = 2 * i * pi / nun_spirals
            t_end = 2 * (i + scale) * pi / nun_spirals
            t = np.linspace(t_start, t_end, size) + np.random.random(size=size) * 0.1
            xs[ix] = np.c_[r * np.sin(t), r * np.cos(t)]
            ys[ix] = i % num_classes
        return cls.from_xy(xs, ys, TaskTypes.CLASSIFICATION)

    @classmethod
    def two_clusters(cls, *,
                     size: int = 100,
                     scale: float = 1.,
                     center: float = 0.,
                     distance: float = 2.,
                     num_dimensions: int = 2) -> "TabularDataset":
        center1 = (np.random.random(num_dimensions) + center - 0.5) * scale + distance
        center2 = (np.random.random(num_dimensions) + center - 0.5) * scale - distance
        cluster1 = (np.random.randn(size, num_dimensions) + center1) * scale
        cluster2 = (np.random.randn(size, num_dimensions) + center2) * scale
        data = np.vstack((cluster1, cluster2)).astype(np_float_type)
        labels = np.array([1] * size + [0] * size, np_int_type)
        indices = np.random.permutation(size * 2)
        data, labels = data[indices], labels[indices]
        return cls.from_xy(data, labels, TaskTypes.CLASSIFICATION)

    @classmethod
    def simple_non_linear(cls, *,
                          size: int = 120) -> "TabularDataset":
        xs = np.random.randn(size, 2).astype(np_float_type) * 1.5
        ys = np.zeros(size, dtype=np_int_type)
        mask = xs[..., 1] >= xs[..., 0] ** 2
        xs[..., 1][mask] += 2
        ys[mask] = 1
        return cls.from_xy(xs, ys, TaskTypes.CLASSIFICATION)

    @classmethod
    def nine_grid(cls, *,
                  size: int = 120) -> "TabularDataset":
        x, y = np.random.randn(2, size).astype(np_float_type)
        labels = np.zeros(size, np_int_type)
        xl, xr = x <= -1, x >= 1
        yf, yc = y <= -1, y >= 1
        x_mid_mask = ~xl & ~xr
        y_mid_mask = ~yf & ~yc
        mask2 = x_mid_mask & y_mid_mask
        labels[mask2] = 2
        labels[(x_mid_mask | y_mid_mask) & ~mask2] = 1
        xs = np.vstack([x, y]).T
        return cls.from_xy(xs, labels, TaskTypes.CLASSIFICATION)

    @staticmethod
    def _fetch_ys(affine_train, affine_test, task_type):
        if task_type is TaskTypes.REGRESSION:
            y_train, y_test = affine_train, affine_test
        else:
            y_train = (affine_train > 0).astype(np_int_type)
            y_test = (affine_test > 0).astype(np_int_type)
        return y_train, y_test

    @classmethod
    def noisy_linear(cls, *,
                     size: int = 10000,
                     n_dim: int = 100,
                     n_valid: int = 5,
                     noise_scale: float = 0.5,
                     task_type: TaskTypes = TaskTypes.REGRESSION,
                     test_ratio: float = 0.15) -> Tuple["TabularDataset", "TabularDataset"]:
        x_train = np.random.randn(size, n_dim)
        x_train_noise = x_train + np.random.randn(size, n_dim) * noise_scale
        x_test = np.random.randn(int(size * test_ratio), n_dim)
        idx = np.random.permutation(n_dim)[:n_valid]
        w = np.random.randn(n_valid, 1)
        affine_train = x_train[..., idx].dot(w)
        affine_test = x_test[..., idx].dot(w)
        tr_set, te_set = map(
            cls.from_xy,
            [x_train_noise, x_test],
            TabularDataset._fetch_ys(affine_train, affine_test, task_type),
            2 * [task_type]
        )
        return tr_set, te_set

    @classmethod
    def noisy_poly(cls, *,
                   p: int = 3,
                   size: int = 10000,
                   n_dim: int = 100,
                   n_valid: int = 5,
                   noise_scale: float = 0.5,
                   task_type: TaskTypes = TaskTypes.REGRESSION,
                   test_ratio: float = 0.15) -> Tuple["TabularDataset", "TabularDataset"]:
        assert p > 1, "p should be greater than 1"
        x_train = np.random.randn(size, n_dim)
        x_train_list = [x_train] + [x_train ** i for i in range(2, p + 1)]
        x_train_noise = x_train + np.random.randn(size, n_dim) * noise_scale
        x_test = np.random.randn(int(size * test_ratio), n_dim)
        x_test_list = [x_test] + [x_test ** i for i in range(2, p + 1)]
        idx_list = [np.random.permutation(n_dim)[:n_valid] for _ in range(p)]
        w_list = [np.random.randn(n_valid, 1) for _ in range(p)]
        o_train = [x[..., idx].dot(w) for x, idx, w in zip(x_train_list, idx_list, w_list)]
        o_test = [x[..., idx].dot(w) for x, idx, w in zip(x_test_list, idx_list, w_list)]
        affine_train, affine_test = map(partial(np.sum, axis=0), [o_train, o_test])
        tr_set, te_set = map(
            cls.from_xy,
            [x_train_noise, x_test],
            TabularDataset._fetch_ys(affine_train, affine_test, task_type),
            2 * [task_type]
        )
        return tr_set, te_set


__all__ = [
    "flat_arr_type", "raw_data_type", "data_type",
    "DataTuple", "ColumnTypes", "TaskTypes", "FeatureInfo", "TabularDataset"
]



import numpy as np

from ..base import Processor


@Processor.register("min_max")
class MinMax(Processor):
    @property
    def input_dim(self) -> int:
        return 1

    @property
    def output_dim(self) -> int:
        return 1

    def fit(self,
            columns: np.ndarray) -> Processor:
        d_min, d_max = columns.min(), columns.max()
        self._caches["min"], self._caches["diff"] = d_min, d_max - d_min
        return self

    def _process(self,
                 columns: np.ndarray) -> np.ndarray:
        d_min, diff = map(self._caches.get, ["min", "diff"])
        columns -= d_min
        columns /= diff
        return columns

    def _recover(self,
                 processed_columns: np.ndarray) -> np.ndarray:
        d_min, diff = map(self._caches.get, ["min", "diff"])
        processed_columns *= diff
        processed_columns += d_min
        return processed_columns


__all__ = ["MinMax"]



from .types import *
from .utils import *
from .wrapper import TabularData


__all__ = [
    "TabularData", "TabularDataset",
    "TaskTypes", "ColumnTypes", "DataTuple",
    "SplitResult", "DataSplitter", "KFold", "KRandom",
    "ImbalancedSampler", "LabelCollators", "DataLoader"
]



import math
import numpy as np

from typing import *
from collections import Counter
from cftool.misc import get_counter_from_arr

from ..types import *
from ...types import *
from ...misc.c import *


class Recognizer:
    def __init__(self,
                 column_name: str,
                 *,
                 is_label: bool = False,
                 task_type: TaskTypes = None,
                 is_valid: Union[bool, None] = None,
                 is_string: Union[bool, None] = None,
                 is_numerical: Union[bool, None] = None,
                 is_categorical: Union[bool, None] = None,
                 numerical_threshold: float = 0.5):
        # is_* : `None` means no information, `False` means 'force not to *', `True` means 'force to *'
        self.name = column_name
        self.is_label = is_label
        self.task_type = task_type
        if is_string is False and is_numerical is False and is_categorical is False:
            if is_valid is None:
                is_valid = False
            elif is_valid:
                raise ValueError(
                    f"column '{column_name}' is neither string, numerical nor categorical, "
                    "but it is still set to be valid"
                )
        self.is_valid = is_valid
        # TODO : check how to use `is_string`
        self.is_string = is_string
        self.is_numerical = is_numerical
        self.is_categorical = is_categorical
        self.numerical_threshold = numerical_threshold
        self._info = self._counter = self._transform_dict = None

    def __str__(self):
        return f"Recognizer({self.info.column_type})"

    __repr__ = __str__

    @property
    def info(self) -> FeatureInfo:
        return self._info

    @property
    def counter(self) -> Counter:
        return self._counter

    @property
    def transform_dict(self) -> Dict[Union[int, str], int]:
        return self._transform_dict

    @property
    def num_unique_values(self) -> Union[int, float]:
        if self._info.is_numerical:
            return math.inf
        return len(self._transform_dict)

    def _make_invalid_info(self,
                           msg: str,
                           contains_nan: bool,
                           nan_mask: np.ndarray) -> "Recognizer":
        self._info = FeatureInfo(contains_nan, None, nan_mask=nan_mask, is_valid=False, msg=msg)
        return self

    @staticmethod
    def _make_string_info(flat_arr, is_valid, msg, unique_values=None) -> FeatureInfo:
        return FeatureInfo(
            False, flat_arr, is_valid=is_valid, msg=msg, need_transform=True,
            column_type=ColumnTypes.STRING, unique_values_sorted_by_counts=unique_values
        )

    def _check_string_column(self,
                             flat_arr: flat_arr_type) -> Tuple[bool, Union[FeatureInfo, None]]:
        if self.is_numerical or self.is_categorical or self.is_string is False:
            return False, None
        all_numeric = is_all_numeric(flat_arr)
        is_regression = self.task_type is TaskTypes.REGRESSION
        if self.is_label and is_regression and not all_numeric:
            raise ValueError("task_type is REGRESSION but labels are not all numeric")
        if all_numeric:
            return False, None
        self._counter = get_counter_from_arr(flat_arr)
        unique_values = [elem[0] for elem in self._counter.most_common()]
        self._transform_dict = {v: i for i, v in enumerate(unique_values)}
        num_unique_values = len(self._transform_dict)
        if not self.is_valid and num_unique_values == 1:
            msg = (f"all values in column {self.name}, which tends to be string column, "
                   "are the SAME. It'll be excluded since it might be redundant")
            return True, self._make_string_info(None, False, msg)
        if not self.is_valid and num_unique_values == len(flat_arr):
            msg = (f"all values in column {self.name}, which tends to be string column, "
                   "are DIFFERENT. It'll be excluded since it might be redundant")
            return True, self._make_string_info(None, False, msg)
        msg = None
        if num_unique_values >= 1e3:
            msg = f"TOO MANY unique values occurred in column {self.name} ({num_unique_values:^12d})"
        return True, self._make_string_info(flat_arr, True, msg, np.array(unique_values))

    def _check_exclude_categorical(self,
                                   num_samples: int,
                                   num_unique_values: int) -> Tuple[str, Union[str, None]]:
        if not self.is_valid and num_unique_values == 1:
            msg = (f"all values in column {self.name}, which tends to be categorical column, "
                   "are the SAME. It'll be excluded since it might be redundant")
            return "exclude", msg
        if not self.is_valid and num_samples == num_unique_values:
            msg = (f"all values in column {self.name}, which tends to be categorical column, "
                   "are DIFFERENT. It'll be excluded since it might be redundant")
            return "exclude", msg
        if not self.is_categorical and num_unique_values >= self.numerical_threshold * num_samples:
            msg = (
                f"TOO MANY unique values occurred in column {self.name} ({num_unique_values:^12d}) "
                "which tends to be categorical column, it'll cast to numerical column to save memory "
                "and possibly for better performance"
            )
            return "numerical", msg
        return "keep", None

    def _generate_categorical_transform_dict(self):
        values = self._info.unique_values_sorted_by_counts.tolist()
        transform_dict = {v: i for i, v in enumerate(values) if not math.isnan(v)}
        if self._info.contains_nan:
            transform_dict["nan"] = len(transform_dict)
        self._transform_dict = transform_dict

    def fit(self,
            flat_arr: flat_arr_type) -> "Recognizer":
        if self.is_valid is False:
            self._info = FeatureInfo(
                None, None, False,
                msg=f"current column ({self.name}) is forced to be invalid"
            )
            return self
        is_string, info = self._check_string_column(flat_arr)
        if is_string:
            self._info = info
            return self
        if isinstance(flat_arr[0], (str, np.str_)):
            np_flat = flat_arr_to_float32(flat_arr)
            if np_float_type != np.float32:
                np_flat = np.asarray(np_flat, np_float_type)
        else:
            np_flat = np.asarray(flat_arr, np_float_type)
        nan_mask = np.isnan(np_flat)
        valid_mask = ~nan_mask
        np_flat_valid = np_flat[valid_mask]
        np_flat_valid_int = np_flat_valid.astype(np_int_type)
        num_samples, num_valid_samples = map(len, [np_flat, np_flat_valid])
        contains_nan = num_samples != num_valid_samples
        if not contains_nan:
            nan_mask = None
        # check whether all nan or not
        if num_valid_samples == 0:
            if self.is_valid:
                np_flat = np.zeros_like(np_flat)
                self._info = FeatureInfo(contains_nan, np_flat, nan_mask=np.zeros_like(nan_mask))
                return self
            msg = (f"all values in column {self.name}, which tends to be categorical column, "
                   "are NaN. It'll be excluded since it might be redundant")
            return self._make_invalid_info(msg, contains_nan, nan_mask)
        # check whether it's a numerical column
        all_int = np.allclose(np_flat_valid, np_flat_valid_int)
        is_classification_label = self.is_label and self.task_type is TaskTypes.CLASSIFICATION
        if (
            self.is_numerical
            or self.is_numerical is None and (
                self.is_categorical is False
                or self.is_label and self.task_type is TaskTypes.REGRESSION
                or self.is_categorical is None and not all_int
            )
        ):
            if not is_classification_label:
                msg, is_valid = None, True
                if self.is_valid is None and np_flat_valid.max() - np_flat_valid.min() <= 1e-8:
                    is_valid = False
                    msg = (f"all values in column {self.name}, which tends to be numerical column, "
                           "are ALL CLOSE. It'll be excluded since it might be redundant")
                self._info = FeatureInfo(
                    contains_nan, np_flat,
                    nan_mask=nan_mask, is_valid=is_valid, msg=msg
                )
                return self
        # deal with categorical column
        np_flat_categorical = np_flat_valid_int if all_int else np_flat_valid
        min_feat = np_flat_categorical.min()
        np_flat_categorical -= min_feat
        max_feature_value = np_flat_categorical.max()
        if not all_int or max_feature_value > 1e6:
            unique_values, counts = np.unique(np_flat_categorical, return_counts=True)
            num_unique_values = len(unique_values)
            need_transform = unique_values[-1] != num_unique_values - 1
        else:
            counts = np.bincount(np_flat_categorical)
            unique_values = np.nonzero(counts)[0]
            num_unique_values = len(unique_values)
            need_transform = num_unique_values != len(counts)
            counts = counts[unique_values]
        need_transform = min_feat != 0 or need_transform
        status, msg = self._check_exclude_categorical(num_samples, num_unique_values)
        if not is_classification_label and status != "keep":
            if status == "numerical":
                self._info = FeatureInfo(contains_nan, np_flat, nan_mask=nan_mask, msg=msg)
                return self
            return self._make_invalid_info(msg, contains_nan, nan_mask)
        sorted_indices = np.argsort(counts)[::-1]
        unique_values = unique_values[sorted_indices] + min_feat
        counter_dict = dict(zip(unique_values, counts))
        if contains_nan:
            unique_values = np.append(unique_values, [float("nan")])
            counter_dict["nan"] = num_samples - num_valid_samples
        need_transform = need_transform or contains_nan
        self._counter = Counter(counter_dict)
        self._info = FeatureInfo(
            contains_nan, np_flat,
            nan_mask=nan_mask,
            need_transform=need_transform,
            column_type=ColumnTypes.CATEGORICAL,
            unique_values_sorted_by_counts=unique_values
        )
        self._generate_categorical_transform_dict()
        return self


__all__ = ["Recognizer"]



import numpy
from Cython.Build import cythonize
from setuptools import setup, find_packages, Extension

VERSION = "0.1.5"

DESCRIPTION = "Data processing module implemented with numpy"
with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="carefree-data",
    version=VERSION,
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        "carefree-toolkit>=0.1.4",
        "dill", "future", "psutil", "pillow",
        "cython>=0.29.12", "numpy>=1.16.2", "scipy>=1.2.1",
        "scikit-learn>=0.20.3", "matplotlib>=3.0.3",
        "mkdocs", "mkdocs-material", "mkdocs-minify-plugin",
        "Pygments", "pymdown-extensions"
    ],
    ext_modules=cythonize(Extension(
        "cfdata.misc.c.cython_utils",
        sources=["cfdata/misc/c/cython_utils.pyx"],
        language="c",
        include_dirs=[numpy.get_include(), "cfdata/misc/c"],
        library_dirs=[],
        libraries=[],
        extra_compile_args=[],
        extra_link_args=[]
    )),
    package_data={"cfdata.misc.c": ["cython_utils.pyx"]},
    author="carefree0910",
    author_email="syameimaru_kurumi@pku.edu.cn",
    url="https://github.com/carefree0910/carefree-data",
    download_url=f"https://github.com/carefree0910/carefree-data/archive/v{VERSION}.tar.gz",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords="python numpy data-science"
)



from typing import Any
from cftool.misc import SavingMixin
from abc import abstractmethod, ABCMeta


class DataBase(SavingMixin, metaclass=ABCMeta):
    @abstractmethod
    def __eq__(self, other: "DataBase"):
        pass

    @abstractmethod
    def read(self, x: Any, y: Any, **kwargs) -> "DataBase":
        pass

    @abstractmethod
    def transform(self, x: Any, y: Any, **kwargs) -> Any:
        pass

    def recover_labels(self, y: Any, *, inplace: bool = False) -> Any:
        pass


__all__ = ["DataBase"]
